import asyncio
import os

from dotbot_neverland import Bot, Context, SelfJoinEvent

bot = Bot()


@bot.on("onlineSet")
async def hello(c: Context[SelfJoinEvent]):
    await c.bot.send("/")


asyncio.run(bot.connect("lounge", "test", os.getenv("HC_PWD"), bypass_dns=True))



# pyright: reportTypedDictNotRequiredAccess = false

from typing import Union

from .models import *
from .utils.parcing_utils import parse_changenick, parse_emote, parse_whisper
from .utils.validators import validate_trip


def parse(
    _data: EventRawData,
) -> Union[
    None,
    Event,
    InfoEvent,
    ChatEvent,
    EmoteEvent,
    WarnEvent,
    SelfJoinEvent,
    CaptchaEvent,
    UserJoinEvent,
    UserLeaveEvent,
    UpdateUserEvent,
    WhisperEvent,
    InviteEvent,
    ChangeNickEvent,
]:
    data = _data.copy()

    if "cmd" not in data:
        return None

    match data["cmd"]:
        case "chat":
            data["trip"] = validate_trip(data.get("trip"))  # type: ignore
            return ChatEvent.parse(data)

        case "warn":
            return WarnEvent.parse(data)

        case "onlineAdd":
            return UserJoinEvent.parse(data, user=User.parse(data))

        case "onlineRemove":
            return UserLeaveEvent.parse(data)

        case "updateUser":
            return UpdateUserEvent.parse(data)

        case "onlineSet":
            users = [User.parse(user) for user in data["users"]]
            data["users"] = users  # type: ignore
            return SelfJoinEvent.parse(data)

        case "captcha":
            return CaptchaEvent.parse(data)

        case "info":
            text_raw = data["text"]

            if "type" not in data:
                data["type"] = None  # type: ignore
                return InfoEvent.parse(data, text_raw=text_raw)

            match data["type"]:
                case "whisper":
                    if isinstance(data["from"], int):
                        # Feedback event, not handled by DotBot-NL
                        return None
                    groups = parse_whisper(text_raw)
                    if groups:
                        data["nick"] = data["from"]
                        data["text"] = groups[1]
                        data["trip"] = validate_trip(data.get("trip"))  # type: ignore
                        return WhisperEvent.parse(data, text_raw=text_raw)
                    else:
                        raise ValueError(
                            f"Failed to parse whisper message: {text_raw!r}"
                        )

                case "emote":
                    groups = parse_emote(text_raw)
                    if groups:
                        (nick, text) = groups
                        data["nick"] = nick
                        data["text"] = text
                        data["trip"] = validate_trip(data.get("trip"))  # type: ignore
                        return EmoteEvent.parse(data, text_raw=text_raw)
                    else:
                        raise ValueError(f"Failed to parse emote message: {text_raw!r}")

                case "invite":
                    if not data["from"]:
                        # Feedback event, not handled by DotBot-NL
                        return None
                    return InviteEvent.parse(
                        data, nick=data["from"], channel=data["inviteChannel"]
                    )

                case _:
                    groups = parse_changenick(text_raw)
                    if groups:
                        return ChangeNickEvent.parse(
                            data,
                            old_nick=groups[0],
                            new_nick=groups[1],
                            text_raw=text_raw,
                        )

                    return InfoEvent.parse(data, text_raw=text_raw)

        case _:
            return None



from typing import Any, Collection, Mapping, cast

from attrs import Attribute, AttrsInstance, fields


def filter_attrs(
    cls: type[AttrsInstance], data: Mapping[str, Any]
) -> Mapping[str, Any]:
    field_names = [
        field.name for field in cast(Collection[Attribute], fields(cls))
    ]  # TODO: cache optimization
    return {k: v for k, v in data.items() if (k in field_names)}



import re

parse_whisper_regex = re.compile(r"(?:You whispered to @.*?: |(.*?) whispered: )(.*)")


def parse_whisper(text: str) -> tuple[str | None, str] | None:
    match = parse_whisper_regex.fullmatch(text)
    if match:
        return match.groups()  # type: ignore

    return None


parse_emote_regex = re.compile(r"(@.*? )(.*)")


def parse_emote(text: str) -> tuple[str, str] | None:
    match = parse_emote_regex.fullmatch(text)
    if match:
        return match.groups()  # type: ignore

    return None


parse_changenick_regex = re.compile(r"(.*) is now (.*)")


def parse_changenick(text: str) -> tuple[str, str] | None:
    match = parse_changenick_regex.match(text)
    if match:
        return match.groups()  # type: ignore

    return None



import asyncio
import os

from dotbot_neverland import Bot, ChatEvent, Context, SelfJoinEvent, WhisperEvent

bot = Bot()


@bot.on("onlineSet")
async def hello(c: Context[SelfJoinEvent]):
    await c.bot.chat("Hello from dotbot-neverland!")


@bot.on("chat", "whisper")
async def ping(c: Context[ChatEvent | WhisperEvent]):
    match c.event.text.split():
        case [":users"]:
            await c.reply(
                "```py\n{}\n```".format(str(list(c.bot.nicks))), whisper_newline=True
            )
        case [":hash", str() as n] if n in c.bot.users_dict:
            await c.reply(c.bot.users_dict[n].hash)
        case [":hash"]:
            await c.reply(c.get_user().hash)


asyncio.run(bot.connect("lounge", "users", os.getenv("HC_PWD"), bypass_dns=True))



import asyncio
import os

from dotbot_neverland import Bot, ChatEvent, Context, SelfJoinEvent

bot = Bot()


@bot.on("onlineSet")
async def hello(c: Context[SelfJoinEvent]):
    await c.bot.chat("Hello from dotbot-neverland!")


@bot.on("chat", ignore_self=True)  # ignore_self is default True, so this can be omitted
async def ping(c: Context[ChatEvent]):
    if c.event.text == ":ping":
        await c.reply(":ping")


asyncio.run(bot.connect("lounge", "ping", os.getenv("HC_PWD"), bypass_dns=True))



from .events import (
    CaptchaEvent,
    ChangeNickEvent,
    ChatEvent,
    EmoteEvent,
    Event,
    InfoEvent,
    InviteEvent,
    SelfJoinEvent,
    UpdateUserEvent,
    UserJoinEvent,
    UserLeaveEvent,
    WarnEvent,
    WhisperEvent,
)
from .raw import EventRawData
from .users import User



import cProfile
import os

import winloop
from dotbot_neverland import Bot, ChatEvent, Context, SelfJoinEvent, WhisperEvent

bot = Bot()


@bot.on("onlineSet")
async def hello(c: Context[SelfJoinEvent]):
    await c.bot.chat("Hello from dotbot-neverland!")


@bot.on("chat", "whisper")
async def ping(c: Context[ChatEvent | WhisperEvent]):
    if c.event.text == ":ping" or c.event.text == "!trip":
        await c.reply("pong!")
    elif c.event.text == ":quit":
        await c.bot.close()


def main():
    winloop.run(bot.connect("lounge", "ping", os.getenv("HC_PWD"), bypass_dns=True))


cProfile.run("main()")



from typing import Literal, Mapping

from ..models.events import *

Alias = Literal[
    "info",
    "chat",
    "emote",
    "warn",
    "onlineSet",
    "captcha",
    "onlineAdd",
    "onlineRemove",
    "updateUser",
    "whisper",
    "invite",
    "changeNick",
]

aliases: Mapping[Alias, type[Event]] = {
    "info": InfoEvent,
    "chat": ChatEvent,
    "emote": EmoteEvent,
    "warn": WarnEvent,
    "onlineSet": SelfJoinEvent,
    "captcha": CaptchaEvent,
    "onlineAdd": UserJoinEvent,
    "onlineRemove": UserLeaveEvent,
    "updateUser": UpdateUserEvent,
    "whisper": WhisperEvent,
    "invite": InviteEvent,
    "changeNick": ChangeNickEvent,
}



import asyncio
import os

from dotbot_neverland import Bot, ChatEvent, Context, SelfJoinEvent, WhisperEvent

bot = Bot()


@bot.on("onlineSet")
async def hello(c: Context[SelfJoinEvent]):
    await c.bot.chat("Hello from dotbot-neverland!")


@bot.on("chat", "whisper")
async def ping(c: Context[ChatEvent | WhisperEvent]):
    if c.event.text == ":ping":
        await c.reply("1")
        await asyncio.sleep(2)
        await c.reply("3")


@bot.on("chat", "whisper")
async def ping2(c: Context[ChatEvent | WhisperEvent]):
    if c.event.text == ":ping":
        await asyncio.sleep(1)
        await c.reply("2")
        await asyncio.sleep(2)
        await c.reply("4")


asyncio.run(bot.connect("lounge", "ping", os.getenv("HC_PWD"), bypass_dns=True))



from .bot import Bot, Context, Listener
from .models import (
    CaptchaEvent,
    ChangeNickEvent,
    ChatEvent,
    EmoteEvent,
    Event,
    EventRawData,
    InfoEvent,
    InviteEvent,
    SelfJoinEvent,
    UpdateUserEvent,
    User,
    UserJoinEvent,
    UserLeaveEvent,
    WarnEvent,
    WhisperEvent,
)



from __future__ import annotations

import asyncio
import json
import ssl
import warnings
from collections import OrderedDict
from typing import Any, Awaitable, Callable, Collection, Literal

import websockets
from attrs import asdict, define

from .models import (
    ChangeNickEvent,
    ChatEvent,
    Event,
    SelfJoinEvent,
    UpdateUserEvent,
    User,
    UserJoinEvent,
    UserLeaveEvent,
    WarnEvent,
    WhisperEvent,
)
from .parsing import parse
from .utils.aliases import Alias, aliases
from .utils.custom_id import generate_customid

type Listener[T: Event] = Callable[[Context[T]], None | Awaitable[None]]
type ListenerKey = type[Event] | Literal["*"]


HC_WS_URL = "wss://hack.chat/chat-ws"


class Bot:
    def __init__(self) -> None:
        self.is_connected = False
        self.listeners: dict[ListenerKey, list[Listener]] = {
            "*": [self._internal_handler]
        }
        self.ws: websockets.WebSocketClientProtocol | None = None
        self.users_dict: OrderedDict[str, User] = OrderedDict()

        self.channel = self.nick = self.password = ""

    async def _connect(self, ws_url: str, ws_opts={}, bypass_dns=False):
        if bypass_dns:
            if ws_url != HC_WS_URL:
                raise ValueError("Bypass DNS is only supported for hack.chat")

            insecure_ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            insecure_ssl_context.check_hostname = False
            insecure_ssl_context.verify_mode = ssl.CERT_NONE
            self.ws = await websockets.connect(
                HC_WS_URL,
                host="104.131.138.176",
                ssl=insecure_ssl_context,
                **ws_opts,
            )
        else:
            self.ws = await websockets.connect(ws_url, **ws_opts)

    async def _join(self, channel: str, nick: str, password: str | None = None):
        payload: dict = {"cmd": "join", "channel": channel, "nick": nick}
        if password:
            payload["password"] = password
        await self.send_json(payload)

    async def connect(
        self,
        channel: str,
        nick: str,
        password: str | None = None,
        ws_url: str = HC_WS_URL,
        ws_opts={},
        bypass_dns=False,
    ):
        self.channel = channel
        self.nick = nick
        self.password = password
        await self._connect(ws_url, ws_opts=ws_opts, bypass_dns=bypass_dns)
        self.is_connected = True
        await self._join(channel, nick, password)
        try:
            async with asyncio.TaskGroup() as tg:
                while self.ws and self.ws.open:
                    message = await self.ws.recv()
                    obj = json.loads(message)
                    event = parse(obj)

                    if event is None:
                        continue

                    ctx = Context(self, event)

                    self.dispatch(tg, "*", ctx)

                    self.dispatch(tg, type(event), ctx)
        except asyncio.exceptions.CancelledError:
            pass
        finally:
            await self.close()

    def dispatch[T: Event](
        self, tg: asyncio.TaskGroup, event_type: type[T] | Literal["*"], ctx: Context[T]
    ):
        listeners = self.listeners.get(event_type, [])

        for listener in listeners:
            result = listener(ctx)

            if asyncio.iscoroutine(result):
                tg.create_task(result)

    async def close(self):
        if self.ws:
            self.is_connected = False
            await self.ws.close()
            self.ws = None

    async def send_raw(self, data: websockets.Data):
        if not self.ws:
            raise RuntimeError("Can't send data when not connected")
        await self.ws.send(data)

    async def send_json(self, obj: Any):
        await self.send_raw(json.dumps(obj))

    async def chat(self, text: str):
        await self.send_json({"cmd": "chat", "text": text})

    async def send(self, text: str):  # the same as `chat`
        await self.send_json({"cmd": "chat", "text": text})

    async def chat_updatable(self, text: str):
        customId = generate_customid()
        await self.send_json({"cmd": "chat", "text": text, "customId": customId})
        return UpdatableMessage(text, customId, self)

    async def whisper(self, nick: str, text: str):
        await self.send_json({"cmd": "whisper", "nick": nick, "text": text})

    async def emote(self, text: str):
        await self.send_json({"cmd": "emote", "text": text})

    async def change_nick(self, nick: str):
        await self.send_json({"cmd": "changenick", "nick": nick})

    async def change_color(self, color: str | None):
        await self.send_json({"cmd": "changecolor", "color": color or "reset"})

    def _add_listener(self, event_type: ListenerKey, listener: Listener):
        if event_type not in self.listeners.keys():
            self.listeners[event_type] = []

        self.listeners[event_type].append(listener)

    def add_listener(
        self,
        matcher: ListenerKey | Alias,
        listener: Listener,
        *,
        ignore_self: bool = True,
    ):
        if isinstance(matcher, str) and matcher != "*":
            matcher = aliases[matcher]
        if ignore_self:

            def _listener(c: Context, /):
                if hasattr(c.event, "nick"):
                    if c.event.nick == self.nick:  # type: ignore
                        return
                return listener(c)
        else:
            _listener = listener
        self._add_listener(matcher, _listener)

    def listen(
        self,
        matchers: Collection[ListenerKey | Alias],
        listener: Listener,
        *,
        ignore_self: bool = True,
    ):
        for matcher in matchers:
            self.add_listener(matcher, listener, ignore_self=ignore_self)

    def on(self, *matchers: ListenerKey | Alias, ignore_self: bool = True):
        def deco(listener: Listener):
            self.listen(matchers, listener, ignore_self=ignore_self)
            return listener

        return deco

    async def _internal_handler(self, ctx: Context[Event]):
        e = ctx.event
        match e:
            case UserJoinEvent():
                self.users_dict[e.user.nick] = e.user
            case UserLeaveEvent():
                self.users_dict.pop(e.nick)
            case SelfJoinEvent():
                self.users_dict = OrderedDict([(user.nick, user) for user in e.users])
            case UpdateUserEvent():
                user = self.users_dict[e.nick]
                # `User.parse` automatically adds `raw` attribute, and `asdict(user)` also provides a `raw`
                # So the previous `raw` must be discarded
                self.users_dict[e.nick] = User.parse({**asdict(user), **e.raw})
            case ChangeNickEvent():
                if e.old_nick == self.nick:
                    self.nick = e.new_nick
            case WarnEvent():
                warnings.warn(f"Warning from server: {e.text}")
        if "channel" in e.raw:
            if (channel := e.raw["channel"]) != self.channel:
                warnings.warn(
                    f"Unexpected channel: {channel} (expecting {self.channel})"
                )

    @property
    def nicks(self):
        return self.users_dict.keys()

    @property
    def users(self):
        return self.users_dict.values()


@define
class Context[T: Event]:
    bot: Bot
    event: T

    @property
    def pair(self):
        return (self.bot, self.event)

    def get_user(self):
        if isinstance(self.event, UserJoinEvent):
            return self.event.user

        elif isinstance(self.event, ChangeNickEvent):
            return self.bot.users_dict[self.event.old_nick]

        return self.bot.users_dict[self.event.nick]  # type: ignore

    def get_channel(self):
        return self.event.raw.get("channel")

    async def reply(self, text: str, whisper_newline: str | bool = False):
        if isinstance(self.event, ChatEvent):
            await self.bot.chat(text)
        elif isinstance(self.event, WhisperEvent):
            if whisper_newline:
                if isinstance(whisper_newline, bool):
                    whisper_newline = ">"
                text = f"{whisper_newline}\n{text}"
            await self.bot.whisper(self.event.nick, text)
        else:
            raise ValueError(
                f"Context[{type(self.event).__name__}] doesn't support reply"
            )


type UpdateMessageMode = Literal["overwrite", "prepend", "append"]


@define
class UpdatableMessage:
    text: str
    customId: str

    bot: Bot

    def _update(self, mode: UpdateMessageMode, text: str):
        return {
            "cmd": "updateMessage",
            "mode": mode,
            "text": text,
            "customId": self.customId,
        }

    def _overwrite(self, text: str):
        return self._update("overwrite", text)

    def _prepend(self, text: str):
        return self._update("prepend", text)

    def _append(self, text: str):
        return self._update("append", text)

    def _complete(self):
        return {
            "cmd": "updateMessage",
            "mode": "complete",
            "customId": self.customId,
        }

    async def update(self, mode: UpdateMessageMode, text: str):
        await self.bot.send_json(self._update(mode, text))

    async def overwrite(self, text: str):
        await self.bot.send_json(self._overwrite(text))

    async def prepend(self, text: str):
        await self.bot.send_json(self._prepend(text))

    async def append(self, text: str):
        await self.bot.send_json(self._append(text))

    async def complete(self):
        await self.bot.send_json(self._complete())



import re

validate_nick_regex = re.compile(r"^[a-zA-Z0-9_]{1,24}$")


def is_valid_nick(nick: str) -> bool:
    return bool(validate_nick_regex.fullmatch(nick))


def strip_nick(nick: str) -> str:
    return nick.strip().removeprefix("@")


def validate_trip(trip: str | None) -> str | None:
    if trip == "" or trip == "null":
        return None
    return trip



import asyncio
import os
import random
import re

from dotbot_neverland import Bot, ChatEvent, Context, SelfJoinEvent

bot = Bot()


@bot.on("onlineSet")
async def hello(c: Context[SelfJoinEvent]):
    await c.bot.change_color("20201D")


say_regex = re.compile(r"说([一-龟]+)")


@bot.on("chat")
async def dpg(c: Context[ChatEvent]):
    if f"@{c.bot.nick} " in c.event.text:
        if match := say_regex.search(c.event.text):
            await c.bot.chat(match.group(1))
        else:
            if random.random() > 0.5:
                await c.bot.chat("ovo")


asyncio.run(bot.connect("lounge", "_", os.getenv("HC_PWD"), bypass_dns=True))



from typing import Self

from attrs import define

from ._base import Base
from .raw import EventRawData
from .users import User

__all__ = [
    "Event",
    "InfoEvent",
    "ChatEvent",
    "EmoteEvent",
    "WarnEvent",
    "SelfJoinEvent",
    "CaptchaEvent",
    "UserJoinEvent",
    "UserLeaveEvent",
    "UpdateUserEvent",
    "WhisperEvent",
    "InviteEvent",
    "ChangeNickEvent",
]


@define
class Event(Base):
    cmd: str
    raw: EventRawData

    @classmethod
    def parse(cls, data: EventRawData, **extra_kwds) -> Self:
        return super().parse(data, raw=data, **extra_kwds)


@define
class InfoEvent(Event):
    type: str | None
    text_raw: str


@define
class ChatEvent(Event):
    nick: str
    trip: str | None
    text: str


@define
class EmoteEvent(InfoEvent):
    nick: str
    trip: str | None
    text: str


@define
class WarnEvent(Event):
    text: str


@define
class SelfJoinEvent(Event):
    users: list[User]


@define
class CaptchaEvent(Event):
    text: str


@define
class UserJoinEvent(Event):
    user: User


@define
class UserLeaveEvent(Event):
    nick: str


@define
class UpdateUserEvent(Event):
    nick: str
    # Some of the other fields are hard to model,
    # because they may or may not be present,
    # and requires complex fallback to something like attrs.NOTHING.
    # Since this event is hardly used, it is acceptable to only use the nick
    # and the raw data.


@define
class WhisperEvent(InfoEvent):
    nick: str
    trip: str | None
    text: str


@define
class InviteEvent(Event):
    nick: str
    channel: str


@define
class ChangeNickEvent(Event):
    old_nick: str
    new_nick: str



from typing import Any, Mapping, TypedDict

EventRawData = TypedDict(
    "EventRawData",
    {
        "cmd": str,
        "channel": str,
        "text": str,
        "type": str,
        "nick": str,
        "trip": str,
        "from": str | int,
        "to": str | int,
        "nicks": list[str],
        "users": list[Mapping[str, Any]],
        "time": str,
        "color": str,
        "admin": bool,
        "mod": bool,
        "inviteChannel": str,
    },
    total=False,
)



from typing import Any, Literal, Mapping, Optional, Self

from attrs import define

from ..utils.validators import validate_trip
from ._base import Base


@define
class User(Base):
    channel: str
    color: Optional[str]
    hash: str
    isBot: bool
    level: int
    nick: str
    trip: Optional[str]
    uType: Literal["user", "mod", "admin"]
    userid: int

    raw: Mapping[str, Any]

    @classmethod
    def parse(cls, _data: Mapping[str, Any], **extra_kwds) -> Self:
        data: dict = _data.copy()  # type: ignore
        data["trip"] = validate_trip(data.get("trip"))
        if "raw" in data:
            data.pop("raw")
        return super().parse(data, raw=data, **extra_kwds)



from typing import Mapping, Self

from attrs import define

from ._filter_attrs import filter_attrs


@define
class Base[T: Mapping]:
    @classmethod
    def parse(cls, _data: T, **extra_kwds) -> Self:
        try:
            return cls(**filter_attrs(cls, _data), **extra_kwds)
        except Exception as e:
            raise ValueError(
                f"Error during parsing\n{_data=}\n{filter_attrs(cls, _data)=}\n{extra_kwds=}"
            ) from e



import asyncio
import os

from dotbot_neverland import Bot, ChatEvent, Context, SelfJoinEvent

bot = Bot()


@bot.on("onlineSet")
async def hello(c: Context[SelfJoinEvent]):
    await c.bot.chat("Hello from dotbot-neverland!")


@bot.on("chat")
async def ping(c: Context[ChatEvent]):
    if c.event.text == ":ping":
        msg = await c.bot.chat_updatable("GO!")
        await asyncio.sleep(1)
        await msg.append("!")
        await asyncio.sleep(1)
        await msg.append("!")
        await asyncio.sleep(1)
        await msg.prepend("My")
        await asyncio.sleep(1)
        await msg.overwrite("BanG Dream! Its MyGO!!!!!")
        await msg.complete()
        # Why so many ppl in HC like MyGO?


asyncio.run(bot.connect("lounge", "ping", os.getenv("HC_PWD"), bypass_dns=True))



import asyncio

from dotbot_neverland import Bot

bot = Bot()


@bot.on("onlineSet")
async def hello(c):
    await c.bot.chat("Hello from dotbot-neverland!")


asyncio.run(bot.connect("lounge", "hello"))
